# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'widget.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QHBoxLayout, QMainWindow, QSizePolicy,
    QTabWidget, QVBoxLayout, QWidget)

class Ui_BrowserWindow(object):
    def setupUi(self, BrowserWindow):
        if not BrowserWindow.objectName():
            BrowserWindow.setObjectName(u"BrowserWindow")
        BrowserWindow.resize(800, 600)
        self.actionCopy = QAction(BrowserWindow)
        self.actionCopy.setObjectName(u"actionCopy")
        self.actionPaste = QAction(BrowserWindow)
        self.actionPaste.setObjectName(u"actionPaste")
        self.actionCut = QAction(BrowserWindow)
        self.actionCut.setObjectName(u"actionCut")
        self.actionUndo = QAction(BrowserWindow)
        self.actionUndo.setObjectName(u"actionUndo")
        self.actionHelp = QAction(BrowserWindow)
        self.actionHelp.setObjectName(u"actionHelp")
        self.actionAdd_new = QAction(BrowserWindow)
        self.actionAdd_new.setObjectName(u"actionAdd_new")
        self.actionAdd_new.setCheckable(False)
        self.actionRemove_current = QAction(BrowserWindow)
        self.actionRemove_current.setObjectName(u"actionRemove_current")
        self.actionEdit_note = QAction(BrowserWindow)
        self.actionEdit_note.setObjectName(u"actionEdit_note")
        self.actionNew_note = QAction(BrowserWindow)
        self.actionNew_note.setObjectName(u"actionNew_note")
        self.actionNew_notefile = QAction(BrowserWindow)
        self.actionNew_notefile.setObjectName(u"actionNew_notefile")
        self.actionNew_notefile.setCheckable(False)
        self.actionNew_notefile.setChecked(False)
        self.actionMake_link = QAction(BrowserWindow)
        self.actionMake_link.setObjectName(u"actionMake_link")
        self.actionNext_notefile = QAction(BrowserWindow)
        self.actionNext_notefile.setObjectName(u"actionNext_notefile")
        self.actionPrevious_notefile = QAction(BrowserWindow)
        self.actionPrevious_notefile.setObjectName(u"actionPrevious_notefile")
        self.actionDelete_selected = QAction(BrowserWindow)
        self.actionDelete_selected.setObjectName(u"actionDelete_selected")
        self.actionZoom_in = QAction(BrowserWindow)
        self.actionZoom_in.setObjectName(u"actionZoom_in")
        self.actionZoom_out = QAction(BrowserWindow)
        self.actionZoom_out.setObjectName(u"actionZoom_out")
        self.actionMake_this_view_point_default_for_the_notefile = QAction(BrowserWindow)
        self.actionMake_this_view_point_default_for_the_notefile.setObjectName(u"actionMake_this_view_point_default_for_the_notefile")
        self.actionMake_this_notefile_appear_first_on_program_start = QAction(BrowserWindow)
        self.actionMake_this_notefile_appear_first_on_program_start.setObjectName(u"actionMake_this_notefile_appear_first_on_program_start")
        self.actionClose = QAction(BrowserWindow)
        self.actionClose.setObjectName(u"actionClose")
        self.actionBulgarian = QAction(BrowserWindow)
        self.actionBulgarian.setObjectName(u"actionBulgarian")
        self.actionEnglish = QAction(BrowserWindow)
        self.actionEnglish.setObjectName(u"actionEnglish")
        self.actionClear_settings_and_exit = QAction(BrowserWindow)
        self.actionClear_settings_and_exit.setObjectName(u"actionClear_settings_and_exit")
        self.actionRename_notefile = QAction(BrowserWindow)
        self.actionRename_notefile.setObjectName(u"actionRename_notefile")
        self.actionDelete_notefile = QAction(BrowserWindow)
        self.actionDelete_notefile.setObjectName(u"actionDelete_notefile")
        self.actionColor_blue = QAction(BrowserWindow)
        self.actionColor_blue.setObjectName(u"actionColor_blue")
        self.actionColor_green = QAction(BrowserWindow)
        self.actionColor_green.setObjectName(u"actionColor_green")
        self.actionColor_red = QAction(BrowserWindow)
        self.actionColor_red.setObjectName(u"actionColor_red")
        self.actionColor_black = QAction(BrowserWindow)
        self.actionColor_black.setObjectName(u"actionColor_black")
        self.actionMove_down = QAction(BrowserWindow)
        self.actionMove_down.setObjectName(u"actionMove_down")
        self.actionMove_up = QAction(BrowserWindow)
        self.actionMove_up.setObjectName(u"actionMove_up")
        self.actionMove_left = QAction(BrowserWindow)
        self.actionMove_left.setObjectName(u"actionMove_left")
        self.actionMove_right = QAction(BrowserWindow)
        self.actionMove_right.setObjectName(u"actionMove_right")
        self.actionSearch = QAction(BrowserWindow)
        self.actionSearch.setObjectName(u"actionSearch")
        self.actionSearch.setCheckable(True)
        self.actionSelect_all_notes = QAction(BrowserWindow)
        self.actionSelect_all_notes.setObjectName(u"actionSelect_all_notes")
        self.actionJump_to_nearest_note = QAction(BrowserWindow)
        self.actionJump_to_nearest_note.setObjectName(u"actionJump_to_nearest_note")
        self.actionDetails = QAction(BrowserWindow)
        self.actionDetails.setObjectName(u"actionDetails")
        self.actionAbout = QAction(BrowserWindow)
        self.actionAbout.setObjectName(u"actionAbout")
        self.actionSelect_note_under_mouse = QAction(BrowserWindow)
        self.actionSelect_note_under_mouse.setObjectName(u"actionSelect_note_under_mouse")
        self.actionCreate_note_from_the_clipboard_text = QAction(BrowserWindow)
        self.actionCreate_note_from_the_clipboard_text.setObjectName(u"actionCreate_note_from_the_clipboard_text")
        self.actionTransparent_background = QAction(BrowserWindow)
        self.actionTransparent_background.setObjectName(u"actionTransparent_background")
        self.actionSet_this_height_as_default = QAction(BrowserWindow)
        self.actionSet_this_height_as_default.setObjectName(u"actionSet_this_height_as_default")
        self.actionCopy_donation_address = QAction(BrowserWindow)
        self.actionCopy_donation_address.setObjectName(u"actionCopy_donation_address")
        self.actionSelect_all_red_notes = QAction(BrowserWindow)
        self.actionSelect_all_red_notes.setObjectName(u"actionSelect_all_red_notes")
        self.actionRedo = QAction(BrowserWindow)
        self.actionRedo.setObjectName(u"actionRedo")
        self.actionCheck_for_updates = QAction(BrowserWindow)
        self.actionCheck_for_updates.setObjectName(u"actionCheck_for_updates")
        self.actionCheck_for_updates.setCheckable(True)
        self.actionCheck_for_updates.setChecked(False)
        self.actionDownload_it = QAction(BrowserWindow)
        self.actionDownload_it.setObjectName(u"actionDownload_it")
        self.actionGotoTab1 = QAction(BrowserWindow)
        self.actionGotoTab1.setObjectName(u"actionGotoTab1")
        self.actionGotoTab2 = QAction(BrowserWindow)
        self.actionGotoTab2.setObjectName(u"actionGotoTab2")
        self.actionSwitch_to_the_last_note_file = QAction(BrowserWindow)
        self.actionSwitch_to_the_last_note_file.setObjectName(u"actionSwitch_to_the_last_note_file")
        self.actionExport_all_as_web_notes = QAction(BrowserWindow)
        self.actionExport_all_as_web_notes.setObjectName(u"actionExport_all_as_web_notes")
        self.actionToggle_tags_view = QAction(BrowserWindow)
        self.actionToggle_tags_view.setObjectName(u"actionToggle_tags_view")
        self.actionToggle_tags_view.setCheckable(True)
        self.actionToggle_tag = QAction(BrowserWindow)
        self.actionToggle_tag.setObjectName(u"actionToggle_tag")
        self.actionUse_JSON_format = QAction(BrowserWindow)
        self.actionUse_JSON_format.setObjectName(u"actionUse_JSON_format")
        self.actionMigrate_to_JSON_format = QAction(BrowserWindow)
        self.actionMigrate_to_JSON_format.setObjectName(u"actionMigrate_to_JSON_format")
        self.centralwidget = QWidget(BrowserWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.horizontalLayout_2 = QHBoxLayout(self.centralwidget)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.mainLayout = QVBoxLayout()
        self.mainLayout.setObjectName(u"mainLayout")
        self.tabBarWidget = QTabWidget(self.centralwidget)
        self.tabBarWidget.setObjectName(u"tabBarWidget")
        self.tabBarWidget.setTabsClosable(True)
        self.tabBarWidget.setMovable(True)

        self.mainLayout.addWidget(self.tabBarWidget)


        self.horizontalLayout_2.addLayout(self.mainLayout)

        BrowserWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(BrowserWindow)
        self.actionClose.triggered.connect(self.actionToggle_tag.toggle)

        QMetaObject.connectSlotsByName(BrowserWindow)
    # setupUi

    def retranslateUi(self, BrowserWindow):
        BrowserWindow.setWindowTitle(QCoreApplication.translate("BrowserWindow", u"MainWindow", None))
        self.actionCopy.setText(QCoreApplication.translate("BrowserWindow", u"&Copy", None))
#if QT_CONFIG(tooltip)
        self.actionCopy.setToolTip(QCoreApplication.translate("BrowserWindow", u"Copies notes to the internal clipboard, and copies their text to the regular clipboard.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionCopy.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+C", None))
#endif // QT_CONFIG(shortcut)
        self.actionPaste.setText(QCoreApplication.translate("BrowserWindow", u"&Paste", None))
#if QT_CONFIG(tooltip)
        self.actionPaste.setToolTip(QCoreApplication.translate("BrowserWindow", u"Pastes the notes from the internal clipboard.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionPaste.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+V", None))
#endif // QT_CONFIG(shortcut)
        self.actionCut.setText(QCoreApplication.translate("BrowserWindow", u"Cut", None))
#if QT_CONFIG(tooltip)
        self.actionCut.setToolTip(QCoreApplication.translate("BrowserWindow", u"See the tooltip of \"Copy\"", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionCut.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+X", None))
#endif // QT_CONFIG(shortcut)
        self.actionUndo.setText(QCoreApplication.translate("BrowserWindow", u"&Undo", None))
#if QT_CONFIG(tooltip)
        self.actionUndo.setToolTip(QCoreApplication.translate("BrowserWindow", u"Reverts the last action. The max is 30 undo steps. (note there is no Redo for now)", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionUndo.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+Z", None))
#endif // QT_CONFIG(shortcut)
        self.actionHelp.setText(QCoreApplication.translate("BrowserWindow", u"&Help", None))
#if QT_CONFIG(tooltip)
        self.actionHelp.setToolTip(QCoreApplication.translate("BrowserWindow", u"Shows the help notefile.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionHelp.setShortcut(QCoreApplication.translate("BrowserWindow", u"F1", None))
#endif // QT_CONFIG(shortcut)
        self.actionAdd_new.setText(QCoreApplication.translate("BrowserWindow", u"&Add another folder", None))
        self.actionRemove_current.setText(QCoreApplication.translate("BrowserWindow", u"&Remove current", None))
#if QT_CONFIG(tooltip)
        self.actionRemove_current.setToolTip(QCoreApplication.translate("BrowserWindow", u"Remove the current folder (does not delete the notefiles from the disk)", None))
#endif // QT_CONFIG(tooltip)
        self.actionEdit_note.setText(QCoreApplication.translate("BrowserWindow", u"&Edit note", None))
#if QT_CONFIG(shortcut)
        self.actionEdit_note.setShortcut(QCoreApplication.translate("BrowserWindow", u"E", None))
#endif // QT_CONFIG(shortcut)
        self.actionNew_note.setText(QCoreApplication.translate("BrowserWindow", u"&New note", None))
#if QT_CONFIG(tooltip)
        self.actionNew_note.setToolTip(QCoreApplication.translate("BrowserWindow", u"Create a new note", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionNew_note.setShortcut(QCoreApplication.translate("BrowserWindow", u"N", None))
#endif // QT_CONFIG(shortcut)
        self.actionNew_notefile.setText(QCoreApplication.translate("BrowserWindow", u"&New notefile", None))
#if QT_CONFIG(tooltip)
        self.actionNew_notefile.setToolTip(QCoreApplication.translate("BrowserWindow", u"Create a new notefile on the disk and open it to add notes.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionNew_notefile.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+N", None))
#endif // QT_CONFIG(shortcut)
        self.actionMake_link.setText(QCoreApplication.translate("BrowserWindow", u"&Make link", None))
#if QT_CONFIG(tooltip)
        self.actionMake_link.setToolTip(QCoreApplication.translate("BrowserWindow", u"Link notes with arrows to visualize their connections.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionMake_link.setShortcut(QCoreApplication.translate("BrowserWindow", u"L", None))
#endif // QT_CONFIG(shortcut)
        self.actionNext_notefile.setText(QCoreApplication.translate("BrowserWindow", u"Next notefile", None))
#if QT_CONFIG(tooltip)
        self.actionNext_notefile.setToolTip(QCoreApplication.translate("BrowserWindow", u"Switch to the next notefile in the alphabetical order.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionNext_notefile.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+PgUp", None))
#endif // QT_CONFIG(shortcut)
        self.actionPrevious_notefile.setText(QCoreApplication.translate("BrowserWindow", u"Previous notefile", None))
#if QT_CONFIG(tooltip)
        self.actionPrevious_notefile.setToolTip(QCoreApplication.translate("BrowserWindow", u"Switch to the previous notefile in thealphabetical order.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionPrevious_notefile.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+PgDown", None))
#endif // QT_CONFIG(shortcut)
        self.actionDelete_selected.setText(QCoreApplication.translate("BrowserWindow", u"&Delete selected", None))
#if QT_CONFIG(tooltip)
        self.actionDelete_selected.setToolTip(QCoreApplication.translate("BrowserWindow", u"Delete selected notes.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionDelete_selected.setShortcut(QCoreApplication.translate("BrowserWindow", u"Del", None))
#endif // QT_CONFIG(shortcut)
        self.actionZoom_in.setText(QCoreApplication.translate("BrowserWindow", u"&Zoom in", None))
#if QT_CONFIG(tooltip)
        self.actionZoom_in.setToolTip(QCoreApplication.translate("BrowserWindow", u"Change the height of the viewpoint over the canvas.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionZoom_in.setShortcut(QCoreApplication.translate("BrowserWindow", u"PgDown", None))
#endif // QT_CONFIG(shortcut)
        self.actionZoom_out.setText(QCoreApplication.translate("BrowserWindow", u"Zoom &out", None))
#if QT_CONFIG(tooltip)
        self.actionZoom_out.setToolTip(QCoreApplication.translate("BrowserWindow", u"Change the height of the viewpoint over the canvas.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionZoom_out.setShortcut(QCoreApplication.translate("BrowserWindow", u"PgUp", None))
#endif // QT_CONFIG(shortcut)
        self.actionMake_this_view_point_default_for_the_notefile.setText(QCoreApplication.translate("BrowserWindow", u"&Make this view point default for the notefile", None))
#if QT_CONFIG(tooltip)
        self.actionMake_this_view_point_default_for_the_notefile.setToolTip(QCoreApplication.translate("BrowserWindow", u"Fix the current viewpoint (height excluded) in the notefile as default on program start.", None))
#endif // QT_CONFIG(tooltip)
        self.actionMake_this_notefile_appear_first_on_program_start.setText(QCoreApplication.translate("BrowserWindow", u"Make this notefile appear first &on program start", None))
#if QT_CONFIG(tooltip)
        self.actionMake_this_notefile_appear_first_on_program_start.setToolTip(QCoreApplication.translate("BrowserWindow", u"Marks the notefile as default on program start for its folder.", None))
#endif // QT_CONFIG(tooltip)
        self.actionClose.setText(QCoreApplication.translate("BrowserWindow", u"Close", None))
#if QT_CONFIG(shortcut)
        self.actionClose.setShortcut(QCoreApplication.translate("BrowserWindow", u"Q", None))
#endif // QT_CONFIG(shortcut)
        self.actionBulgarian.setText(QCoreApplication.translate("BrowserWindow", u"\u0411\u044a\u043b\u0433\u0430\u0440\u0441\u043a\u0438", None))
        self.actionEnglish.setText(QCoreApplication.translate("BrowserWindow", u"&English", None))
#if QT_CONFIG(tooltip)
        self.actionEnglish.setToolTip(QCoreApplication.translate("BrowserWindow", u"Change the language to english.", None))
#endif // QT_CONFIG(tooltip)
        self.actionClear_settings_and_exit.setText(QCoreApplication.translate("BrowserWindow", u"&Clear settings and exit", None))
#if QT_CONFIG(tooltip)
        self.actionClear_settings_and_exit.setToolTip(QCoreApplication.translate("BrowserWindow", u"Clear program settings and kill this instance.", None))
#endif // QT_CONFIG(tooltip)
        self.actionRename_notefile.setText(QCoreApplication.translate("BrowserWindow", u"&Rename notefile", None))
#if QT_CONFIG(tooltip)
        self.actionRename_notefile.setToolTip(QCoreApplication.translate("BrowserWindow", u"Also changes all of the notes that link to it to the new name.", None))
#endif // QT_CONFIG(tooltip)
        self.actionDelete_notefile.setText(QCoreApplication.translate("BrowserWindow", u"&Delete notefile", None))
#if QT_CONFIG(tooltip)
        self.actionDelete_notefile.setToolTip(QCoreApplication.translate("BrowserWindow", u"Delete the notefile permanetly (on the disk too).", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionDelete_notefile.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+Del", None))
#endif // QT_CONFIG(shortcut)
        self.actionColor_blue.setText(QCoreApplication.translate("BrowserWindow", u"&Color blue", None))
#if QT_CONFIG(shortcut)
        self.actionColor_blue.setShortcut(QCoreApplication.translate("BrowserWindow", u"1", None))
#endif // QT_CONFIG(shortcut)
        self.actionColor_green.setText(QCoreApplication.translate("BrowserWindow", u"Color &green", None))
#if QT_CONFIG(shortcut)
        self.actionColor_green.setShortcut(QCoreApplication.translate("BrowserWindow", u"2", None))
#endif // QT_CONFIG(shortcut)
        self.actionColor_red.setText(QCoreApplication.translate("BrowserWindow", u"Color &red", None))
#if QT_CONFIG(shortcut)
        self.actionColor_red.setShortcut(QCoreApplication.translate("BrowserWindow", u"3", None))
#endif // QT_CONFIG(shortcut)
        self.actionColor_black.setText(QCoreApplication.translate("BrowserWindow", u"Color &black", None))
#if QT_CONFIG(shortcut)
        self.actionColor_black.setShortcut(QCoreApplication.translate("BrowserWindow", u"4", None))
#endif // QT_CONFIG(shortcut)
        self.actionMove_down.setText(QCoreApplication.translate("BrowserWindow", u"&Move down", None))
#if QT_CONFIG(shortcut)
        self.actionMove_down.setShortcut(QCoreApplication.translate("BrowserWindow", u"Down", None))
#endif // QT_CONFIG(shortcut)
        self.actionMove_up.setText(QCoreApplication.translate("BrowserWindow", u"Move &up", None))
#if QT_CONFIG(shortcut)
        self.actionMove_up.setShortcut(QCoreApplication.translate("BrowserWindow", u"Up", None))
#endif // QT_CONFIG(shortcut)
        self.actionMove_left.setText(QCoreApplication.translate("BrowserWindow", u"Move &left", None))
#if QT_CONFIG(shortcut)
        self.actionMove_left.setShortcut(QCoreApplication.translate("BrowserWindow", u"Left", None))
#endif // QT_CONFIG(shortcut)
        self.actionMove_right.setText(QCoreApplication.translate("BrowserWindow", u"Move &right", None))
#if QT_CONFIG(shortcut)
        self.actionMove_right.setShortcut(QCoreApplication.translate("BrowserWindow", u"Right", None))
#endif // QT_CONFIG(shortcut)
        self.actionSearch.setText(QCoreApplication.translate("BrowserWindow", u"&Search ", None))
#if QT_CONFIG(tooltip)
        self.actionSearch.setToolTip(QCoreApplication.translate("BrowserWindow", u"Search in all notes. Click in the results shown under the search field to go to the note.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionSearch.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+F", None))
#endif // QT_CONFIG(shortcut)
        self.actionSelect_all_notes.setText(QCoreApplication.translate("BrowserWindow", u"S&elect all notes", None))
#if QT_CONFIG(shortcut)
        self.actionSelect_all_notes.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+A", None))
#endif // QT_CONFIG(shortcut)
        self.actionJump_to_nearest_note.setText(QCoreApplication.translate("BrowserWindow", u"&Jump to nearest note", None))
#if QT_CONFIG(shortcut)
        self.actionJump_to_nearest_note.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+J", None))
#endif // QT_CONFIG(shortcut)
        self.actionDetails.setText(QCoreApplication.translate("BrowserWindow", u"Details", None))
#if QT_CONFIG(shortcut)
        self.actionDetails.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+D", None))
#endif // QT_CONFIG(shortcut)
        self.actionAbout.setText(QCoreApplication.translate("BrowserWindow", u"&About", None))
        self.actionSelect_note_under_mouse.setText(QCoreApplication.translate("BrowserWindow", u"Select note under mouse", None))
#if QT_CONFIG(shortcut)
        self.actionSelect_note_under_mouse.setShortcut(QCoreApplication.translate("BrowserWindow", u"Return", None))
#endif // QT_CONFIG(shortcut)
        self.actionCreate_note_from_the_clipboard_text.setText(QCoreApplication.translate("BrowserWindow", u"Create &note from the clipboard text", None))
#if QT_CONFIG(tooltip)
        self.actionCreate_note_from_the_clipboard_text.setToolTip(QCoreApplication.translate("BrowserWindow", u"Create note from the text in the regular clipboard.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionCreate_note_from_the_clipboard_text.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+Shift+V", None))
#endif // QT_CONFIG(shortcut)
        self.actionTransparent_background.setText(QCoreApplication.translate("BrowserWindow", u"&Transparent background", None))
#if QT_CONFIG(tooltip)
        self.actionTransparent_background.setToolTip(QCoreApplication.translate("BrowserWindow", u"Changes only the background color of the note to be transparent.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionTransparent_background.setShortcut(QCoreApplication.translate("BrowserWindow", u"5", None))
#endif // QT_CONFIG(shortcut)
        self.actionSet_this_height_as_default.setText(QCoreApplication.translate("BrowserWindow", u"&Set this height as default", None))
#if QT_CONFIG(tooltip)
        self.actionSet_this_height_as_default.setToolTip(QCoreApplication.translate("BrowserWindow", u"Set the current height as default for all files on startup.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(statustip)
        self.actionSet_this_height_as_default.setStatusTip(QCoreApplication.translate("BrowserWindow", u"Set the current height as default for all files on startup.", None))
#endif // QT_CONFIG(statustip)
#if QT_CONFIG(whatsthis)
        self.actionSet_this_height_as_default.setWhatsThis(QCoreApplication.translate("BrowserWindow", u"Set the current height as default for all files on startup.", None))
#endif // QT_CONFIG(whatsthis)
        self.actionCopy_donation_address.setText(QCoreApplication.translate("BrowserWindow", u"&Copy donation address", None))
        self.actionSelect_all_red_notes.setText(QCoreApplication.translate("BrowserWindow", u"Se&lect all red notes", None))
#if QT_CONFIG(shortcut)
        self.actionSelect_all_red_notes.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+R", None))
#endif // QT_CONFIG(shortcut)
        self.actionRedo.setText(QCoreApplication.translate("BrowserWindow", u"&Redo", None))
#if QT_CONFIG(shortcut)
        self.actionRedo.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+Shift+Z", None))
#endif // QT_CONFIG(shortcut)
        self.actionCheck_for_updates.setText(QCoreApplication.translate("BrowserWindow", u"Check for updates", None))
        self.actionDownload_it.setText(QCoreApplication.translate("BrowserWindow", u"Download it", None))
        self.actionGotoTab1.setText(QCoreApplication.translate("BrowserWindow", u"gotoTab1", None))
#if QT_CONFIG(shortcut)
        self.actionGotoTab1.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+1", None))
#endif // QT_CONFIG(shortcut)
        self.actionGotoTab2.setText(QCoreApplication.translate("BrowserWindow", u"gotoTab2", None))
#if QT_CONFIG(shortcut)
        self.actionGotoTab2.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+2", None))
#endif // QT_CONFIG(shortcut)
        self.actionSwitch_to_the_last_note_file.setText(QCoreApplication.translate("BrowserWindow", u"S&witch to the last note file", None))
#if QT_CONFIG(shortcut)
        self.actionSwitch_to_the_last_note_file.setShortcut(QCoreApplication.translate("BrowserWindow", u"Backspace", None))
#endif // QT_CONFIG(shortcut)
        self.actionExport_all_as_web_notes.setText(QCoreApplication.translate("BrowserWindow", u"&Export all as web notes", None))
        self.actionToggle_tags_view.setText(QCoreApplication.translate("BrowserWindow", u"Toggle tags view", None))
#if QT_CONFIG(shortcut)
        self.actionToggle_tags_view.setShortcut(QCoreApplication.translate("BrowserWindow", u"Ctrl+T", None))
#endif // QT_CONFIG(shortcut)
        self.actionToggle_tag.setText(QCoreApplication.translate("BrowserWindow", u"Toggle tag", None))
#if QT_CONFIG(shortcut)
        self.actionToggle_tag.setShortcut(QCoreApplication.translate("BrowserWindow", u"T", None))
#endif // QT_CONFIG(shortcut)
        self.actionUse_JSON_format.setText(QCoreApplication.translate("BrowserWindow", u"Use JSON format", None))
        self.actionMigrate_to_JSON_format.setText(QCoreApplication.translate("BrowserWindow", u"Migrate to JSON format", None))
    # retranslateUi




from PySide6.QtWidgets import QWidget
from .ui_widget import Ui_AnchorEditPropsWidget


class AnchorEditPropsWidget(QWidget):
    def __init__(self, parent):
        super().__init__(parent)

        self.ui = Ui_AnchorEditPropsWidget()
        self.ui.setupUi(self)



from __future__ import annotations
import hashlib
from pathlib import Path

from PySide6.QtGui import QImageReader, QImage
from PySide6.QtCore import QMimeData
from fusion.libs.entity import Entity
from fusion.util.point2d import Point2D

import pamet
from pamet.desktop_app.util import jpeg_blob_from_image
from pamet.model.card_note import CardNote
from pamet.model.image_note import ImageNote
from pamet.model.text_note import TextNote
from pamet.util import snap_to_grid
from pamet.constants import ALIGNMENT_GRID_UNIT
from pamet.util.url import Url
from pamet.views.note.qt_helpers import minimal_nonelided_size
from fusion import get_logger

log = get_logger(__name__)


def entities_from_mime_data(mime_data: QMimeData) -> list[Entity]:
    entities = []

    if mime_data.hasImage():
        try:
            image = mime_data.imageData()
            if image.isNull():
                raise Exception(f'Could not load image from clipboard')

            blob = jpeg_blob_from_image(image)
            local_path = pamet.desktop_app.media_store().save_blob(
                blob, 'jpg', cache_subfolder_name='__pasted_images__')
        except Exception as e:
            log.error(f'Failed to save image from clipboard: {e}')
            return []

        image_note = ImageNote()
        image_note.local_image_url = local_path
        image_note.image_size = Point2D(image.size().width(),
                                        image.size().height())

        image_note.image_md5 = hashlib.md5(blob).hexdigest()

        entities.append(image_note)

    elif mime_data.hasUrls():
        # urls: list[str] = mime_data.text().split('\n')
        for url in mime_data.urls():
            # Check if it's a local file
            as_local_path = Path(url.toLocalFile())
            if as_local_path.exists():

                # Check if it's an image
                suffix = as_local_path.suffix.lower()
                suffix = suffix[1:] if suffix else suffix
                if (suffix in QImageReader.supportedImageFormats()):
                    note = ImageNote()
                    # Read to qimage
                    image_data = as_local_path.read_bytes()
                    image = QImage.fromData(image_data)
                    if image.isNull():
                        log.error(f'Could not load image from {as_local_path}')
                        continue

                    # Get md5
                    md5sum = hashlib.md5(image_data).hexdigest()

                    note.local_image_url = str(as_local_path)
                    note.image_size = Point2D(image.size().width(),
                                              image.size().height())
                    note.image_md5 = md5sum

                else:
                    # TODO: Support other file types
                    continue

            else:
                note = TextNote()
                note.text = url

                parsed_url = Url(url)
                if parsed_url.has_web_schema():
                    note.url = url

            entities.append(note)

    elif mime_data.hasText():  # This triggers for URLs too
        text = mime_data.text()
        text_sections = text.split('\n\n')

        for section in text_sections:
            note = TextNote()
            note.text = section
            # Single urls don't get detected by Qt
            if Url(section).has_web_schema():
                note.url = section
            entities.append(note)

    # Autosize the notes and place them one under the other
    next_spawn_pos = Point2D(0, 0)
    for note in entities:
        rect = note.rect()

        # Autosize
        if isinstance(note, (TextNote, CardNote, ImageNote)):
            new_size = minimal_nonelided_size(note)
            rect.set_size(snap_to_grid(new_size))

        # Move under the last one
        rect.set_top_left(next_spawn_pos)

        # Update the spawn pos to be one unit under the last note
        next_spawn_pos = rect.bottom_left()
        next_spawn_pos += Point2D(0, ALIGNMENT_GRID_UNIT)

        # Insert the note
        note.set_rect(rect)

    return entities



from fusion.libs.entity import entity_type
from pamet.model.text_note import TextNote


@entity_type
class OtherPageListNote(TextNote):
    @property
    def text(self) -> str:
        return '*Other page links*'

    @text.setter
    def text(self, new_text: str):
        raise Exception



from importlib import resources

from PySide6.QtGui import QFont, QFontDatabase
from fusion.extensions_loader import ExtensionsLoader
from fusion.platform.qt_widgets import configure_for_qt as fusion_config_qt
from pamet.desktop_app.config import UserDesktopSettings
from pamet.desktop_app.config import pamet_data_folder_path
from pamet.desktop_app.util import copy_script_templates
from pamet.services.media_store import MediaStore
from pamet import desktop_app
from fusion.logging import get_logger
from pamet.util import resource_path

log = get_logger(__name__)


def configure_for_qt(app):
    global _media_store, _default_note_font

    log.info(f'Using data folder: {pamet_data_folder_path}')
    desktop_app.set_app(app)
    fusion_config_qt(app)

    config: UserDesktopSettings = desktop_app.get_user_settings()
    if config.changes_present():
        desktop_app.save_user_settings(config)

    copy_script_templates()

    desktop_app.set_media_store(MediaStore(config.media_store_path))
    desktop_app.icons.load_all()

    _font_id = QFontDatabase.addApplicationFont(
        str(resource_path('fonts/OpenSans-VariableFont_wdth,wght.ttf')))
    _font_family = QFontDatabase.applicationFontFamilies(_font_id)[0]
    _default_note_font = QFont(_font_family)
    _default_note_font.setPointSizeF(14)
    desktop_app.set_default_note_font(_default_note_font)

    views_dir = resources.files('pamet') / 'views'
    pamet_root = views_dir.parent
    views_loader = ExtensionsLoader(pamet_root)
    views_loader.load_all_recursively(views_dir)



from copy import copy
from PySide6.QtCore import QPoint, QPointF, QRectF, QSizeF, Qt
from PySide6.QtGui import QPainter, QPainterPath, QPalette, QPolygon
from fusion.util.point2d import Point2D

from fusion.libs.state import view_state_type
from pamet import desktop_app
from pamet.desktop_app.util import draw_text_lines
from pamet.model.script_note import ScriptNote
from pamet.note_view_lib import register_note_view_type
from pamet.views.note.base.state import NoteViewState
from pamet.views.note.text.widget import TextNoteWidget


@view_state_type
class ScriptNoteViewState(NoteViewState, ScriptNote):
    pass


@register_note_view_type(state_type=ScriptNoteViewState,
                         note_type=ScriptNote,
                         edit=False)
class ScriptNoteWidget(TextNoteWidget):

    def __init__(self, parent, initial_state):
        TextNoteWidget.__init__(self, parent, initial_state)

    def left_mouse_double_click_event(self, position: Point2D):
        desktop_app.script_runner.run(self.state().get_note())

    def paintEvent(self, event):
        painter = QPainter()
        painter.begin(self)

        draw_text_lines(painter, self._elided_text_layout.data,
                        self._alignment,
                        self.state().text_rect())

        # Draw the link decorations
        pen = painter.pen()

        internal_border_rect = QRectF(self.rect())
        internal_border_rect.setSize(internal_border_rect.size() -
                                     QSizeF(1, 1))
        internal_border_rect.moveTopLeft(QPointF(0.5, 0.5))

        # Draw the external link decoration
        # Draw a triangular marker in the upper left corner
        top_right = self.rect().topRight()
        p1 = copy(top_right)
        p1.setX(p1.x() - 10)
        p2 = copy(p1)
        p2.setY(p1.y() + 10)
        p3 = copy(p1)
        p3.setX(top_right.x())
        p3.setY(p1.y() + 5)

        offset = QPoint(-3, 3)
        p1 += offset
        p2 += offset
        p3 += offset

        poly = QPolygon()
        poly << p1 << p2 << p3

        path = QPainterPath()
        path.addPolygon(poly)
        brush = painter.brush()
        brush.setStyle(Qt.SolidPattern)
        brush.setColor(self.palette().color(QPalette.WindowText))
        painter.fillPath(path, brush)

        # Draw the solid border
        painter.drawRect(internal_border_rect)

        painter.end()



from PySide6.QtCore import QRect
from PySide6.QtGui import QColor, QPaintEvent, QPainter
from fusion.libs.entity.change import Change
from fusion.platform.qt_widgets import bind_and_apply_state
from fusion.libs.state import view_state_type
from fusion.logging import get_logger

from pamet import register_note_view_type
from pamet.model.image_note import ImageNote
from pamet.views.note.base.view import NoteView
from pamet.views.note.base.state import NoteViewState
from pamet.views.note.image.image_label import ImageLabel
from pamet.views.note.qt_helpers import draw_link_decorations

log = get_logger(__name__)


@view_state_type
class ImageNoteViewState(NoteViewState, ImageNote):
    pass


@register_note_view_type(state_type=ImageNoteViewState,
                         note_type=ImageNote,
                         edit=False)
class ImageNoteWidget(ImageLabel, NoteView):

    def __init__(self, parent, initial_state):
        ImageLabel.__init__(self, parent)
        NoteView.__init__(self, parent, initial_state)

        bind_and_apply_state(self, initial_state, self.on_state_change)

    def on_state_change(self, change: Change):
        state: ImageNoteViewState = change.last_state()

        if change.updated.color or change.updated.background_color:
            fg_col = QColor(*state.get_color().to_uint8_rgba_list())
            bg_col = QColor(*state.get_background_color().to_uint8_rgba_list())

            palette = self.palette()
            palette.setColor(self.backgroundRole(), bg_col)
            palette.setColor(self.foregroundRole(), fg_col)
            self.setPalette(palette)

        if change.updated.local_image_url:
            self.update_image_cache(state.image_url, state.local_image_url)
        if change.updated.geometry:
            self.setGeometry(QRect(*state.geometry))

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter()
        draw_link_decorations(self, painter)
        super().paintEvent(event)



from fusion.libs.entity.change import Change
from fusion.libs.channel import Subscription
from pamet import channels
import pamet
from pamet.actions import note as note_actions
from pamet.desktop_app.util import elide_text
from pamet.util.url import Url


class LinkNoteViewMixin:

    def __init__(self):
        self._page_subscription: Subscription = None

    @property
    def displayed_text(self):
        if not self._elided_text_layout:
            return '!! Text not processed !!'
        return self._elided_text_layout.text()

    def recalculate_elided_text(self):
        state = self.state()
        url: Url = state.url
        if url.is_internal():
            page = pamet.page(url.get_page_id())
            if page:
                self._elided_text_layout = elide_text(page.name,
                                                      state.text_rect(),
                                                      self.font())
            else:
                self._elided_text_layout = elide_text('(missing)',
                                                      state.text_rect(),
                                                      self.font())
        else:  # Is not internal
            self._elided_text_layout = elide_text(state.text,
                                                  state.text_rect(),
                                                  self.font())

    def connect_to_page_changes(self, page_id):
        if self._page_subscription:
            self.disconnect_from_page_changes()

        self._page_subscription = channels.entity_changes_by_id.subscribe(
            self.handle_page_change, index_val=page_id)

    def disconnect_from_page_changes(self):
        if self._page_subscription:
            self._page_subscription.unsubscribe()
            self._page_subscription = None

    def handle_page_change(self, change: Change):
        if change.is_delete() or change.updated.name:
            note_actions.apply_page_change_to_anchor_view(self.state())



from PySide6.QtCore import QRect
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QMessageBox
from pamet.model.op_list_note import OtherPageListNote
from pamet.note_view_lib import register_note_view_type
from pamet.views.note.op_list.state import OtherPageListNoteViewState
from pamet.views.note.text.widget import TextNoteWidget


@register_note_view_type(state_type=OtherPageListNoteViewState,
                         note_type=OtherPageListNote,
                         edit=False)
class OtherPageListNoteWidget(TextNoteWidget):
    def left_mouse_double_click_event(self, position):
        QMessageBox.information(self, 'Other page list info', '''
        The OtherPageList note cannot be edited. It's function is to keep an
        up-to-date list with links to all pages in the Pamet repo. It's useful
        for keeping an updated index of your pages.
        ''')

    def on_state_change(self, change):
        state = change.last_state()
        if change.updated.color or change.updated.background_color:

            fg_col = QColor(*state.get_color().to_uint8_rgba_list())
            bg_col = QColor(*state.get_background_color().to_uint8_rgba_list())

            palette = self.palette()
            palette.setColor(self.backgroundRole(), bg_col)
            palette.setColor(self.foregroundRole(), fg_col)
            self.setPalette(palette)

        if change.updated.geometry:
            self.setGeometry(QRect(*state.geometry))

        if change.updated.text or \
                change.updated.geometry \
                or change.updated.url:
            self.recalculate_elided_text()



from fusion.storage.in_memory_repository import InMemoryRepository
from pamet.model.arrow import Arrow
from pamet.model.note import Note
from pamet.model.page import Page
from pamet.storage.base_repository import PametRepository


class PametInMemoryRepository(InMemoryRepository, PametRepository):
    def __init__(self):
        InMemoryRepository.__init__(self, (Page, Note, Arrow))


import subprocess

from PySide6.QtWidgets import QWidget
from PySide6.QtGui import QColor, QImage, QPainter, QBrush
from PySide6.QtGui import QCursor
from PySide6.QtCore import Qt, QRect, QPoint, QTimer, Signal
from PySide6.QtWidgets import QApplication


class SelectorWidget(QWidget):
    # Signals
    image_selected = Signal(QImage)

    def __init__(self):

        super(SelectorWidget, self).__init__()

        self.setWindowFlags(Qt.FramelessWindowHint)
        self.setAttribute(Qt.WA_TranslucentBackground)

        self.dragging = False
        self.taking_screenshot = False
        self.mousePressPoint = QPoint()
        self.selectedRect = QRect()

    def finish(self):
        self.taking_screenshot = False
        self.close()

    def keyPressEvent(self, event):
        self.finish()

    def rectFromAnyTwoPoints(self, p1, p2):
        x1 = min(p1.x(), p2.x())
        y1 = min(p1.y(), p2.y())
        w = max(p1.x(), p2.x()) - x1
        h = max(p1.y(), p2.y()) - y1

        return QRect(x1, y1, w, h)

    def grabSelectedRect(self):
        screen = self.window().windowHandle().screen()
        pixmap = screen.grabWindow(0)

        if not self.selectedRect.isValid():
            print('Selected rect is not valid')
            self.finish()
            return

        region = pixmap.copy(self.selectedRect)

        # This has to be before setting the clipboard contents
        # because otherwise the notfy-send hangs
        # subprocess.run(['notify-send', 'Copied snippet to clipboard'])

        clipboard = QApplication.clipboard()
        clipboard.setImage(region.toImage())

        self.finish()

    def mousePressEvent(self, event):
        self.mousePressPoint = event.pos()
        self.dragging = True

    def mouseReleaseEvent(self, event):
        self.selectedRect = self.rectFromAnyTwoPoints(
            self.mousePressPoint, event.pos())

        self.taking_screenshot = True
        self.dragging = False
        self.update()
        QTimer.singleShot(100, self.grabSelectedRect)  # Wait for the DM update

    def mouseMoveEvent(self, event):
        self.update()

    def paintEvent(self, event):
        painter = QPainter()

        painter.begin(self)

        if self.taking_screenshot:
            painter.setBrush(QBrush(QColor(250, 250, 250, 0), Qt.SolidPattern))
            painter.setPen(Qt.NoPen)
            painter.drawRect(self.rect())

            return

        painter.setBrush(QBrush(QColor(100, 100, 100, 40), Qt.CrossPattern))
        painter.setPen(Qt.NoPen)

        painter.drawRect(self.rect())

        painter.setPen(Qt.red)
        painter.setBrush(QBrush(QColor(100, 100, 100, 0), Qt.SolidPattern))

        if self.dragging:
            p1 = self.mousePressPoint
            p2 = self.mapFromGlobal(QCursor.pos())

            painter.drawRect(self.rectFromAnyTwoPoints(p1, p2))

        painter.end()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'properties_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QFrame, QHBoxLayout,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_MapPagePropertiesWidget(object):
    def setupUi(self, MapPagePropertiesWidget):
        if not MapPagePropertiesWidget.objectName():
            MapPagePropertiesWidget.setObjectName(u"MapPagePropertiesWidget")
        MapPagePropertiesWidget.resize(335, 503)
        self.verticalLayout = QVBoxLayout(MapPagePropertiesWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label = QLabel(MapPagePropertiesWidget)
        self.label.setObjectName(u"label")

        self.verticalLayout.addWidget(self.label)

        self.nameLineEdit = QLineEdit(MapPagePropertiesWidget)
        self.nameLineEdit.setObjectName(u"nameLineEdit")

        self.verticalLayout.addWidget(self.nameLineEdit)

        self.nameWarningLabel = QLabel(MapPagePropertiesWidget)
        self.nameWarningLabel.setObjectName(u"nameWarningLabel")

        self.verticalLayout.addWidget(self.nameWarningLabel)

        self.saveButton = QPushButton(MapPagePropertiesWidget)
        self.saveButton.setObjectName(u"saveButton")

        self.verticalLayout.addWidget(self.saveButton)

        self.line = QFrame(MapPagePropertiesWidget)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.verticalLayout.addWidget(self.line)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(-1, 10, -1, -1)
        self.setAsHomePageCheckBox = QCheckBox(MapPagePropertiesWidget)
        self.setAsHomePageCheckBox.setObjectName(u"setAsHomePageCheckBox")

        self.horizontalLayout_2.addWidget(self.setAsHomePageCheckBox)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.verticalSpacer = QSpacerItem(20, 274, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.deleteButton = QPushButton(MapPagePropertiesWidget)
        self.deleteButton.setObjectName(u"deleteButton")

        self.verticalLayout.addWidget(self.deleteButton)


        self.retranslateUi(MapPagePropertiesWidget)

        QMetaObject.connectSlotsByName(MapPagePropertiesWidget)
    # setupUi

    def retranslateUi(self, MapPagePropertiesWidget):
        MapPagePropertiesWidget.setWindowTitle(QCoreApplication.translate("MapPagePropertiesWidget", u"Form", None))
        self.label.setText(QCoreApplication.translate("MapPagePropertiesWidget", u"Name:", None))
        self.nameWarningLabel.setText(QCoreApplication.translate("MapPagePropertiesWidget", u"You should not be seeing this, lol", None))
        self.saveButton.setText(QCoreApplication.translate("MapPagePropertiesWidget", u"Save (Ctrl+S)", None))
#if QT_CONFIG(shortcut)
        self.saveButton.setShortcut(QCoreApplication.translate("MapPagePropertiesWidget", u"Ctrl+S", None))
#endif // QT_CONFIG(shortcut)
        self.setAsHomePageCheckBox.setText(QCoreApplication.translate("MapPagePropertiesWidget", u"Set as home page when starting the app", None))
        self.deleteButton.setText(QCoreApplication.translate("MapPagePropertiesWidget", u"Delete page", None))
    # retranslateUi




import time
from typing import Dict, Generator, List
import os
import json
from pathlib import Path
from pamet import desktop_app
from pamet.desktop_app import get_repo_settings, get_user_settings
from pamet.model.arrow import Arrow
from pamet.model.page_child import PageChild
from pamet.storage.pamet_in_memory_repo import PametInMemoryRepository
from slugify import slugify

import fusion
from fusion.libs.entity import Entity, dump_to_dict, load_from_dict
from fusion.libs.entity.change import Change
from fusion.libs.channel import Channel
from fusion.storage.in_memory_repository import InMemoryRepository
from fusion import get_logger

import pamet
from pamet.model.page import Page
from pamet.model.note import Note

from .legacy import LegacyFSRepoReader

log = get_logger(__name__)

V4_FILE_EXT = '.pam4.json'


class FSStorageRepository(PametInMemoryRepository, LegacyFSRepoReader):
    """File system storage. This class has all entities cached at all times"""

    def __init__(self, path, queue_save_on_change=False):
        PametInMemoryRepository.__init__(self)
        LegacyFSRepoReader.__init__(self)

        self.path = Path(path)
        self.queue_save_on_change = queue_save_on_change
        self._page_ids = []

        self._save_channel: Channel = None

        self.page_paths_by_id: Dict[str, Path] = {}

        self.upserted_pages = set()
        self.removed_pages = set()

    def load_all_pages(self):
        # Load all pages in the cache
        for page_path in self.page_paths():
            try:
                entities = self.get_entities_from_json(
                    page_path)  #@IgnoreException
                if not entities:
                    continue
                page, notes, arrows = entities
            except Exception as e:
                log.error(
                    f'Exception raised while loading page {page_path}: {e}')
                continue

            # Check if the page is saved under the right name
            inferred_path = self.path_for_page(page)
            if page_path != inferred_path:  # If not - fix it by re-saving
                page_path.unlink()
                self.create_page_on_disk(page, notes, arrows)
                log.error(f'Bad page name. Deleted {page_path} and saved the '
                          f'page {page.name} anew.')

            # Check for dupicates
            page_duplicates = list(self.find(gid=page.gid()))
            for page_duplicate in page_duplicates:
                dup_path = self.path_for_page(page_duplicate)

                if dup_path.stat().st_mtime <= page_path.stat().st_mtime:
                    # If the page that's older is loaded
                    InMemoryRepository.remove_one(self, page_duplicate)
                    backup_path = dup_path.with_suffix('.backup')
                    dup_path.rename(backup_path)
                    log.error(f'Found duplicate page "{page_duplicate.name} '
                              f'in the repo. Backed it up as {backup_path}')
                else:
                    # If the page, currently processed in the upper loop
                    # is older
                    backup_path = page_path.with_suffix('.backup')
                    dup_path.rename(backup_path)
                    log.error(f'Found duplicate page "{page_duplicate.name} '
                              f'in the repo. Backed it up as {backup_path}')
                    break

            # Add the entities to the in-memory cache
            try:
                InMemoryRepository.insert_one(self, page)
                for note in notes:
                    try:
                        InMemoryRepository.insert_one(self, note)
                    except Exception:
                        log.error(f'Duplicate note. Skipping {note}')
                        continue
                for arrow in arrows:
                    try:
                        InMemoryRepository.insert_one(self, arrow)
                    except Exception:
                        log.error(f'Duplicate arrow. Skipping {arrow}')
                        continue
            except Exception:
                log.error(f'Duplicate page. Skipping {page}')
                continue

            # Save the path corresponding to the id in order to handle renames
            self.page_paths_by_id[page.id] = self.path_for_page(page)

    def path_for_page(self, page) -> Path:
        slug = slugify(page.name, separator='_', max_length=100)
        name = f'{slug}-{page.id}{V4_FILE_EXT}'
        return self.path / name

    def id_from_page_path(self, path: Path) -> str:
        # Remove the file ext .pam4.json and get the file name
        page_id = path.with_suffix('').stem

        slug, page_id = page_id.split('-')
        return page_id

    @classmethod
    def open(cls, path, **kwargs):
        if not os.path.exists(path) or not os.path.isdir(path):
            raise Exception('Bad path. Cannot create repository for', path)

        return cls(path, **kwargs)

    @classmethod
    def new(cls, path, **kwargs):
        if os.path.exists(path):
            if os.listdir(path):
                raise Exception(
                    f'Cannot create repository in non-empty folder {path}')

        os.makedirs(path, exist_ok=True)
        return cls(path, **kwargs)

    def set_save_channel(self, save_channel):
        self._save_channel = save_channel

    def insert_one(self, entity: Entity):
        if isinstance(entity, Page):
            self.upserted_pages.add(entity)
        elif isinstance(entity, PageChild):
            page = self.page(entity.page_id)
            if not page:
                raise Exception(f'Invalid parent for {entity}')
            self.upserted_pages.add(page)

        InMemoryRepository.insert_one(self, entity)

        return Change.CREATE(entity)

    def insert(self, entities: List[Entity]):
        for entity in entities:
            self.insert_one(entity)

    def remove_one(self, entity):
        InMemoryRepository.remove_one(self, entity)
        if isinstance(entity, Page):
            self.removed_pages.add(entity)
        elif isinstance(entity, PageChild):
            page = self.page(entity.page_id)
            if not page:
                return None
            self.upserted_pages.add(page)

        return Change.DELETE(entity)

    def remove(self, entities: List[Entity]):
        for entity in entities:
            self.remove_one(entity)

    def update_one(self, entity):
        change = InMemoryRepository.update_one(self, entity)
        if isinstance(entity, Page):
            self.upserted_pages.add(entity)
        elif isinstance(entity, PageChild):
            page = self.page(entity.page_id)
            if not page:
                raise Exception(f'Invalid parent for {entity}')
            self.upserted_pages.add(page)
        return change

    def update(self, entities: List[Entity]):
        for entity in entities:
            self.update_one(entity)

    def find(self, **filter):
        yield from InMemoryRepository.find(self, **filter)

    def _try_to_save(self, changes: List[Change]):
        try:
            for change in changes:
                if change.is_create():
                    self.insert_one(change.last_state())
                elif change.is_delete():
                    self.remove_one(change.last_state())
                elif change.is_update():
                    self.update_one(change.last_state())

            self.write_to_disk(changes)
        except Exception as e:
            # Here I could handle exceptions silently, retry, show in GUI, etc
            raise e
        # else:
        #     if self._save_channel:
        #         self._save_channel.push(changes)

    def write_to_disk(self):
        if self.upserted_pages.intersection(self.removed_pages):
            raise Exception('A page is both marked for upsert and removal.')

        for page in self.upserted_pages:
            page = pamet.page(page.id)

            if page.id in self.page_paths_by_id:
                self.update_page_on_disk(page, pamet.notes(page),
                                         pamet.arrows(page))
            else:
                self.create_page_on_disk(page, pamet.notes(page),
                                         pamet.arrows(page))

        for page in self.removed_pages:
            self.delete_page_on_disk(page)

        self.upserted_pages.clear()
        self.removed_pages.clear()

    def page_paths(self) -> Generator[Path, None, None]:
        for file in os.scandir(self.path):
            if self.is_v4_page(file.path):
                yield Path(file.path)

    def get_entities_from_json(self, json_file_path):
        try:
            with open(json_file_path) as pf:
                page_state = json.load(pf)  #@IgnoreException
        except Exception as e:
            log.error('Exception %s while loading page' % e, json_file_path)
            return None

        # # TODO REMOVE
        # page_state.pop('type_name', None)
        # if 'note_states' in page_state:
        #     page_state['notes'] = page_state.pop('note_states', [])
        # if 'arrow_states' in page_state:
        #     page_state['arrows'] = page_state.pop('arrow_states', [])
        if 'type_name' not in page_state:
            page_state['type_name'] = Page.__name__

        # Detach the children
        note_states = page_state.pop('notes', [])
        arrow_states = page_state.pop('arrows', [])

        # Create the page
        page = load_from_dict(page_state)

        # Load the notes
        notes = []
        for ns in note_states:
            # Refactoring fixes .. REMOVE LATER
            # if 'x' in ns:
            #     x = ns.pop('x')
            #     y = ns.pop('y')
            #     width = ns.pop('width')
            #     height = ns.pop('height')
            #     ns['geometry'] = [x, y, width, height]

            # if 'time_created' in ns:
            #     ns['created'] = ns.pop('time_created')
            #     ns['modified'] = ns.pop('time_modified')

            # if ns['type_name'] == 'AnchorNote':
            #     ns['type_name'] = 'TextNote'

            id = ns['id']
            if isinstance(id, str):
                assert 'page_id' in ns
                ns['id'] = (ns.pop('page_id'), id)

            if 'script_args_str' in ns:
                command_args = ns.pop('script_args_str')
                if 'content' not in ns:
                    ns['content'] = {}
                ns['content']['command_args'] = command_args

            content = ns.get('content', None)
            if content:

                if 'script' in content:
                    script_path = content.pop('script')
                    content['script_path'] = script_path

            if 'color' in ns:
                for prop in ['color', 'background_color']:
                    prop_val = ns.pop(prop)
                    if 'style' not in ns:
                        ns['style'] = {}
                    ns['style'][prop] = prop_val

            # /ad-hoc fixes

            notes.append(load_from_dict(ns))

        # Load the arrows
        arrows = []
        for arrow_state in arrow_states:
            # REMOVE vvv
            id = arrow_state['id']
            if isinstance(id, str):
                assert 'page_id' in arrow_state
                arrow_state['id'] = (arrow_state.pop('page_id'), id)

            if 'type_name' not in arrow_state:
                arrow_state['type_name'] = Arrow.__name__
            # arrow_state.pop('background_color', None)
            # if 'mid_point_coords' not in arrow_state:
            #     arrow_state['mid_point_coords'] = arrow_state.pop('
            # mid_points')
            # if 'head_coords' not in arrow_state:
            #     arrow_state['head_coords'] = arrow_state.pop('head_point')
            # if 'tail_coords' not in arrow_state:
            #     arrow_state['tail_coords'] = arrow_state.pop('tail_point')

            # if 'page_id' not in arrow_state or not arrow_state['page_id']:
            #     raise Exception

            try:
                arrow: Arrow = load_from_dict(arrow_state)
            except Exception as e:
                log.error(f'Exception {e} raised while parsing arrow '
                          f'{arrow_state} from file {json_file_path}')
                continue

            arrows.append(arrow)

        return page, notes, arrows

    @staticmethod
    def serialize_page(page: Page, notes: List[Note], arrows: List[Arrow]):
        page_state = dump_to_dict(page)
        page_state['notes'] = [dump_to_dict(n) for n in notes]
        page_state['arrows'] = [dump_to_dict(a) for a in arrows]

        try:
            json_str = json.dumps(page_state, ensure_ascii=False, indent=4)
        except Exception as e:
            raise e  # Or log error
            return None

        return json_str

    def create_page_on_disk(self, page: Page, notes: List[Note],
                            arrows: List[Arrow]):
        path = self.path_for_page(page)
        try:
            if path.exists():
                log.error('Cannot create page. File already exists %s' % path)
                return

            page_json_str = self.serialize_page(page, notes, arrows)
            if not page_json_str:
                return

            with open(path, 'w') as pf:
                pf.write(page_json_str)

            self.page_paths_by_id[page.id] = path
        except Exception as e:
            log.error('Exception while writing page at %s: %s' % (path, e))

        return path

    def update_page_on_disk(self, page: Page, notes: List[Note],
                            arrows: List[Arrow]):
        saved_path = self.page_paths_by_id[page.id]
        path = self.path_for_page(page)

        # If the paths differ - the page has been renamed and the file should
        # be moved
        if saved_path != path and saved_path.exists():
            log.debug(f'Page renamed, moving file {saved_path} to {path}.')
            if path.exists():
                raise Exception('This should have been handlet at init time')
            saved_path.rename(path)
            self.page_paths_by_id[page.id] = path

        page_json_str = self.serialize_page(page, notes, arrows)
        if not page_json_str:
            return

        with open(path, 'w') as pf:
            pf.write(page_json_str)

    def delete_page_on_disk(self, page):
        path = self.page_paths_by_id.pop(page.id)

        if os.path.exists(path):
            os.remove(path)

        else:
            log.error('Cannot delete missing page: %s' % path)

    def is_v4_page(self, file_path: str | Path):
        file_path = Path(file_path)
        if file_path.name.endswith(V4_FILE_EXT):
            return True

        return False

    def catch_legacy_version(self, file_path):
        fname = os.path.basename(file_path)
        name, ext = os.path.splitext(fname)
        version = 0

        if ext == 'json':
            version = 3
        elif ext == 'misl':
            version = 1

        return version

    def save_changes(self, changes: List[Change]):
        t0 = time.time()
        self._try_to_save(changes)

        # If there's a big save lag log a warining
        if time.time() - t0 > 0.03:
            log.warning('The save time was above 30ms. Maybe it\'s time to '
                        'implement the async IO')

    def default_page(self):
        repo_settings = get_repo_settings(self.path)
        default_page = self.page(repo_settings.home_page)
        return default_page

    def set_default_page(self, new_page: Page):
        repo_settings = get_repo_settings(self.path)
        repo_settings.home_page = new_page.id
        desktop_app.save_repo_settings(repo_settings)



from copy import copy
from PySide6.QtCore import QPointF, QRectF, QSizeF, Qt
from PySide6.QtGui import QPainter, QPainterPath, QPalette, QPolygon
from fusion.util.point2d import Point2D
from fusion.util.rectangle import Rectangle

import pamet
from pamet.constants import ALIGNMENT_GRID_UNIT, DEFAULT_NOTE_HEIGHT
from pamet.constants import DEFAULT_NOTE_WIDTH, MAX_NOTE_WIDTH
from pamet.constants import MIN_NOTE_HEIGHT, MIN_NOTE_WIDTH
from pamet.constants import PREFERRED_TEXT_NOTE_ASPECT_RATIO
from pamet.desktop_app import default_note_font
from pamet.desktop_app.util import TextLayout, elide_text
from pamet.model.card_note import CardNote
from pamet.model.image_note import ImageNote
from pamet.model.text_note import TextNote


def minimal_nonelided_size(note: TextNote | CardNote | ImageNote) -> Point2D:
    """Do a binary search to get the minimal note size"""
    default_note_size = Point2D(DEFAULT_NOTE_WIDTH, DEFAULT_NOTE_HEIGHT)
    note_font = default_note_font()

    # If it's an image note - fit to the image (if no image - default size)
    if type(note) == ImageNote:  # Card inherits Image!, don't use isinstance
        if note.image_size:
            return note.image_rect()
        else:
            return default_note_size

    if note.url.is_internal():
        internal_page = pamet.page(note.url.get_page_id())
        if internal_page:
            text = str(internal_page.name)
        else:
            text = '(missing)'  # TODO: This is not DRY
    else:
        text = note.text

    if not text:
        return default_note_size

    # Start with the largest possible rect
    # test_note = state.get_note()
    max_w = MAX_NOTE_WIDTH

    unit = ALIGNMENT_GRID_UNIT
    min_width_u = int(MIN_NOTE_WIDTH / unit)
    min_height_u = int(MIN_NOTE_HEIGHT / unit)

    # Do a binary search for the proper width (keeping the aspect ratio)
    low_width_bound = 0
    high_width_bound = round(max_w / unit - min_width_u)
    while (high_width_bound - low_width_bound) > 0:
        test_width_it = (high_width_bound + low_width_bound) // 2
        test_width_u = min_width_u + test_width_it
        test_height_u = round(test_width_u / PREFERRED_TEXT_NOTE_ASPECT_RATIO)

        # test_note.set_rect(
        #     Rectangle(0, 0, test_width_u * unit, test_height_u * unit))
        test_size = Point2D(test_width_u * unit, test_height_u * unit)
        text_layout = elide_text(text, note.text_rect(test_size), note_font)
        if text_layout.is_elided:
            low_width_bound = test_width_it + 1
        else:
            high_width_bound = test_width_it

    # Fine adjust the size by reducing it one unit per step and
    # stopping upon text elide
    width_u = min_width_u + low_width_bound
    height_u = round(width_u / PREFERRED_TEXT_NOTE_ASPECT_RATIO)
    width = width_u * unit
    height = height_u * unit

    # Adjust the height
    rect = Rectangle(0, 0, width, height)
    text_layout = TextLayout()
    while rect.width() >= MIN_NOTE_WIDTH and rect.height() >= MIN_NOTE_HEIGHT:
        if text_layout.is_elided:
            break
        else:
            height = rect.height()

        rect.set_height(rect.height() - unit)
        text_layout = elide_text(text, note.text_rect(rect.size()), note_font)

    # Adjust the width. We check for changes in the text, because
    # even elided text (if it's multi line) can have empty space laterally
    text_layout = elide_text(text, note.text_rect(Point2D(width, height)),
                             note_font)
    text_before_adjust = text_layout.text()
    text = text_before_adjust
    rect = Rectangle(0, 0, width, height)
    while rect.width() >= MIN_NOTE_WIDTH and rect.height() >= MIN_NOTE_HEIGHT:
        if text != text_before_adjust:
            break
        else:
            width = rect.width()

        rect.set_width(rect.width() - unit)
        # test_note.set_rect(rect)
        text_layout = elide_text(text, note.text_rect(rect.size()), note_font)
        text = text_layout.text()

    return Point2D(width, height)


def draw_link_decorations(note_widget, painter: QPainter):
    state = note_widget.state()

    # Draw the link decorations
    pen = painter.pen()
    url: pamet.util.url.Url = state.url

    internal_border_rect = QRectF(note_widget.rect())
    internal_border_rect.setSize(internal_border_rect.size() - QSizeF(1, 1))
    internal_border_rect.moveTopLeft(QPointF(0.5, 0.5))

    if url.is_internal():
        # For internal notes draw only a solid border
        # But if the target page is missing - draw it with a dashed pattern
        if not pamet.page(url.get_page_id()):
            pen.setStyle(Qt.DashLine)
            painter.setPen(pen)
        painter.drawRect(internal_border_rect)

    elif url.is_external():
        # Draw the external link decoration
        # Draw a triangular marker in the upper left corner
        p1 = note_widget.rect().topRight()
        p2 = copy(p1)
        p1.setX(p1.x() - 10)
        p2.setY(p2.y() + 10)

        poly = QPolygon()
        poly << p1 << p2 << note_widget.rect().topRight()

        path = QPainterPath()
        path.addPolygon(poly)
        brush = painter.brush()
        brush.setStyle(Qt.SolidPattern)
        brush.setColor(note_widget.palette().color(QPalette.WindowText))
        painter.fillPath(path, brush)

        # Draw the solid border
        painter.drawRect(internal_border_rect)


from PySide6.QtCore import Qt

from PySide6.QtWidgets import QMainWindow, QPushButton, QTabBar, QWidget
from PySide6.QtGui import QIcon, QKeySequence, QMouseEvent, QResizeEvent, QShortcut

from fusion.libs.entity.change import Change
from fusion.platform.qt_widgets import bind_and_apply_state
from fusion.view import View
from pamet import commands
from pamet.views.tab.widget import TabWidget
from pamet.views.command_palette.widget import CommandPaletteViewState
from pamet.views.command_palette.widget import CommandPaletteWidget
from pamet.views.window.corner_widget.widget import CornerWidget
from pamet.views.window.ui_widget import Ui_BrowserWindow
from pamet.actions import window as window_actions
from pamet.actions import tab as tab_actions

WINDOW_MIN_WIDTH = 400
WINDOW_MIN_HEIGHT = 400
COMMAND_VIEW_SIDE_SPACING = 50
COMMAND_VIEW_MIN_WIDTH = WINDOW_MIN_WIDTH - 2 * COMMAND_VIEW_SIDE_SPACING
COMMAND_VIEW_MAX_WIDTH = 600


class TabBar(QTabBar):  # Adds a tab close on middle-click

    def mouseReleaseEvent(self, event: QMouseEvent):
        if event.button() == Qt.MiddleButton:
            self.tabCloseRequested.emit(self.tabAt(event.pos()))
        return super().mouseReleaseEvent(event)


class WindowWidget(QMainWindow, View):

    def __init__(self, initial_state):
        QMainWindow.__init__(self)
        View.__init__(self, initial_state)

        self.ui = Ui_BrowserWindow()
        self.ui.setupUi(self)
        self.setMinimumWidth(WINDOW_MIN_WIDTH)
        self.setMinimumHeight(WINDOW_MIN_HEIGHT)

        self.tab_widgets = {}
        self.command_widget: QWidget = None

        self.menuButton = QPushButton(QIcon.fromTheme('menu'), '')
        self.corner_widget = CornerWidget(self)

        self.ui.tabBarWidget.setTabBar(TabBar())
        self.ui.tabBarWidget.setTabsClosable(True)
        self.ui.tabBarWidget.setCornerWidget(self.corner_widget)

        self.ui.tabBarWidget.currentChanged.connect(self.handle_tab_changed)
        self.ui.tabBarWidget.tabCloseRequested.connect(
            self.handle_tab_close_requested)

        QShortcut(QKeySequence('ctrl+shift+P'), self,
                  commands.open_command_palette)
        QShortcut(QKeySequence('ctrl+shift+F'), self,
                  commands.show_global_search)
        QShortcut(QKeySequence(Qt.Key_Escape), self, self.handle_esc_shorcut)
        QShortcut(QKeySequence('ctrl+w'), self, commands.close_current_tab)

        go_to_file_shortcut = QShortcut(QKeySequence('ctrl+P'), self)
        go_to_file_shortcut.activated.connect(
            commands.open_command_palette_go_to_file)

        bind_and_apply_state(self, initial_state, self.on_state_change)

    def on_state_change(self, change: Change):
        if change.updated.title:
            self.setWindowTitle(change.last_state().title)

        for tab_state in change.added.tab_states:
            tab = TabWidget(parent=self, initial_state=tab_state)
            self.ui.tabBarWidget.addTab(tab, tab_state.title)
            self.tab_widgets[tab_state.view_id] = tab

        for tab_state in change.removed.tab_states:
            tab = self.tab_widgets[tab_state.view_id]
            tab_idx = self.ui.tabBarWidget.indexOf(tab)
            self.ui.tabBarWidget.removeTab(tab_idx)
            tab.deleteLater()

        if change.updated.command_view_state:
            if self.command_widget:
                self.command_widget.deleteLater()
                self.command_widget = None

            command_view_state = change.last_state().command_view_state
            if isinstance(command_view_state, CommandPaletteViewState):
                self.command_widget = CommandPaletteWidget(
                    parent=self, initial_state=command_view_state)
                self.command_widget.show()
                self._update_command_widget_geometry()

    def handle_tab_changed(self, index: int):
        # If there's no tab currently
        if index == -1:
            self.close()
            return

        self.ui.tabBarWidget.widget(index).update()

    def current_tab(self) -> TabWidget:
        return self.ui.tabBarWidget.currentWidget()

    def _update_command_widget_geometry(self):
        if not self.command_widget:
            return

        # Fixed width
        width = self.geometry().size().width()
        command_palette_width = width - 2 * COMMAND_VIEW_SIDE_SPACING
        command_palette_width = min(COMMAND_VIEW_MAX_WIDTH,
                                    command_palette_width)
        command_palette_width = max(COMMAND_VIEW_MIN_WIDTH,
                                    command_palette_width)

        # Max height 40% от self
        command_palette_height = 0.4 * self.geometry().size().height()
        cw_rect = self.command_widget.geometry()

        # Set the size and position centered in the top of the window
        cw_size = cw_rect.size()
        cw_size.setWidth(command_palette_width)
        cw_size.setHeight(command_palette_height)
        cw_rect.setSize(cw_size)

        cw_rect.moveCenter(self.geometry().center())
        cw_rect.moveLeft(width / 2 - command_palette_width / 2)
        cw_rect.moveTop(0)
        self.command_widget.setGeometry(cw_rect)
        self.command_widget.setMaximumHeight(command_palette_height)
        self.command_widget.setMaximumWidth(command_palette_width)
        self.command_widget.setMinimumWidth(command_palette_width)

    def resizeEvent(self, event: QResizeEvent) -> None:
        self._update_command_widget_geometry()

    def handle_esc_shorcut(self):
        # We're handling the escape shortcut actions here, since the shortcut
        # context management in Qt doesn't allow overloading (or there's a bug)

        state = self.state()
        # Close the command palette if open
        if state.command_view_state:
            window_actions.close_command_view(state)

        # Propagate to the map page if there's one
        current_tab = self.current_tab()
        if not current_tab:
            return

        tab_state = current_tab.state()
        # Close sidebars if open
        if tab_state.right_sidebar_is_open():
            tab_actions.close_page_properties(tab_state)
        if tab_state.left_sidebar_is_open():
            tab_actions.close_left_sidebar(tab_state)

        map_page = current_tab.page_view()
        if not map_page:
            return

        map_page.handle_esc_shortcut()

    def handle_tab_close_requested(self, tab_index: int):
        tab_widget = self.ui.tabBarWidget.widget(tab_index)
        window_actions.close_tab(self.state(), tab_widget.state())



from __future__ import annotations

from pathlib import Path
from typing import Union, Tuple
from urllib.parse import urlparse, urlunparse

from fusion.logging import get_logger
from fusion.util import Point2D

import pamet
from pamet.constants import DEFAULT_EYE_HEIGHT

log = get_logger(__name__)


class Url:

    def __init__(self, url: Union[Url, str]):
        if isinstance(url, Url):
            url = str(url)
        self._url: str = url
        self._parsed_url = urlparse(url)

    def __repr__(self) -> str:
        return urlunparse(self._parsed_url)

    def __eq__(self, other: Url) -> bool:
        return self._url == other._url

    def is_internal(self):
        return self._parsed_url.scheme == 'pamet'

    def is_external(self):
        if not str(self._url):
            return False
        return self._parsed_url.scheme in ['http', 'https', '']

    def has_web_schema(self):
        return self._parsed_url.scheme in ['http', 'https']

    def is_custom(self):
        return self._parsed_url.scheme not in ['pamet', 'http', 'https', '']

    def get_page_id(self):
        # This method should probably not be in this class
        if not self.is_internal():
            # A search for imported public/shared pages should be done here
            return None

        path = Path(self._parsed_url.path)
        parts = path.parts
        if not parts[1] == 'p' or len(parts) < 3:
            log.warning('Invalid url')
            return None

        return parts[2]

    def get_media_id(self):
        path = Path(self._parsed_url.path)
        parts = path.parts
        if not parts[1] == 'p' or not parts[3] == 'media' or len(parts) < 5:
            log.warning('Requested media id for url that has none')
            return None

        return parts[4]

    def get_anchor(self) -> Note | Tuple[float, Point2D] | None:
        fragment = self._parsed_url.fragment
        if fragment.startswith('note'):
            try:
                note_own_id = fragment.split('=')[1]
            except Exception:
                return None
            return pamet.note(self.get_page_id(), own_id=note_own_id)
        elif fragment.startswith('eye_at'):
            try:
                center_tuple = fragment.split('=')[1]
                data = center_tuple.split('/')
                data = [float(c) for c in data]
                height = data[0]
                center = Point2D(data[1], data[2])
                return height, center
            except Exception:
                return None

    def with_anchor(self,
                    anchor: str = None,
                    eye_height: float = None,
                    eye_pos: Point2D = None,
                    note: Note = None) -> Url:
        eye_is_specified = bool(eye_height is not None or eye_pos)
        if bool(anchor) + eye_is_specified + bool(note) != 1:
            raise Exception(
                'Specifiy either eye_at, note or the anchor string')

        if eye_is_specified:
            if eye_height is None:
                eye_height = DEFAULT_EYE_HEIGHT
            elif eye_pos is None:
                eye_pos = Point2D(0, 0)
            anchor = f'eye_at={eye_height}/{eye_pos.x()}/{eye_pos.y()}'

        if note:
            anchor = f'note={note.own_id}'

        new_url = Url(self)
        new_url._parsed_url = new_url._parsed_url._replace(fragment=anchor)
        return new_url

    def is_empty(self):
        return not self._url

    @property
    def hostname(self) -> Url:
        return self._parsed_url.hostname

    @property
    def netloc(self) -> Url:
        return self._parsed_url.netloc

    @property
    def path(self) -> Url:
        return self._parsed_url.path

    @property
    def scheme(self) -> Url:
        return self._parsed_url.scheme



from dataclasses import field
from typing import List

from fusion.libs.state import view_state_type, ViewState
from pamet.model.note import Note
from pamet.views.map_page.state import MapPageViewState


@view_state_type
class TabViewState(ViewState):
    title: str = ''
    page_view_state: MapPageViewState = field(default=None)
    note_edit_view_state: ViewState = None
    # search_bar_state: ViewState = None
    creating_note: Note = None

    left_sidebar_state: ViewState = None
    page_edit_view_state: ViewState = None

    navigation_history: List[str] = field(default_factory=list)
    current_nav_index: int = None
    previous_nav_index: int = None

    page_state_cache: dict = field(default_factory=dict)

    def __repr__(self) -> str:
        return (f'<TabViewState title={self.title}>')

    def set_navigation_index(self, new_index: int):
        new_index = max(0, min(len(self.navigation_history) - 1, new_index))

        if new_index == self.current_nav_index:
            return

        self.previous_nav_index = self.current_nav_index
        self.current_nav_index = new_index

    def right_sidebar_is_open(self):
        return bool(self.page_edit_view_state)

    def left_sidebar_is_open(self):
        return bool(self.left_sidebar_state)

    def add_page_state_to_cache(self, page_state: MapPageViewState):
        if page_state.id in self.page_state_cache:
            return
        self.page_state_cache[page_state.page_id] = page_state

    def page_state_from_cache(self, page_id: str) -> MapPageViewState:
        return self.page_state_cache.get(page_id, None)



from __future__ import annotations

from fusion import fsm
from fusion.libs.action import action
from fusion.util import Point2D
from fusion.logging import get_logger

import pamet
from pamet.constants import DEFAULT_EYE_HEIGHT, MAX_NAVIGATION_HISTORY
from pamet.model.note import Note
from pamet.model.text_note import TextNote
from pamet.views.arrow.widget import ArrowViewState
from pamet.views.map_page.state import MapPageViewState
from pamet.model.page import Page
from pamet.util import generate_page_name
from pamet.util.url import Url
from pamet.actions import map_page as map_page_actions
from pamet.actions import note as note_actions
from pamet.views.search_bar.widget import SearchBarWidgetState
from pamet.views.semantic_search_bar.widget import SemanticSearchBarWidgetState

log = get_logger(__name__)


@action('create_and_set_page')
def get_or_create_page_view(tab_state: TabViewState, page: Page):
    if not page:
        raise Exception('No page at the given URL')

    page_view_state = tab_state.page_state_from_cache(page.id)

    if not page_view_state:
        # Create the page state with all of its children states
        page_view_state = MapPageViewState(id=page.id)
        fsm.add_state(page_view_state)

        for _note in pamet.notes(page):
            NoteViewType = pamet.note_view_type(
                note_type_name=type(_note).__name__)
            StateType = pamet.note_state_type_by_view(NoteViewType.__name__)
            note_props = _note.asdict()
            note_view_state = StateType(**note_props)
            page_view_state.note_view_states.add(note_view_state)
            fsm.add_state(note_view_state)

        for arrow in pamet.arrows(page):
            arrow_props = arrow.asdict()
            arrow_view_state = ArrowViewState(**arrow_props)
            page_view_state.arrow_view_states.add(arrow_view_state)
            fsm.add_state(arrow_view_state)
        fsm.update_state(page_view_state)

    tab_state.add_page_state_to_cache(page_view_state)
    fsm.update_state(tab_state)

    return page_view_state


def set_page_view_state(tab_state: TabViewState,
                        page_view_state: MapPageViewState):

    tab_state.page_view_state = page_view_state
    tab_state.title = pamet.page(page_view_state.page_id).name
    fsm.update_state(tab_state)


@action('update_current_url')
def update_current_url(tab_state: TabViewState):
    if tab_state.current_nav_index is None:
        return

    new_url = tab_state.page_view_state.page_url()
    tab_state.navigation_history[tab_state.current_nav_index] = new_url
    fsm.update_state(tab_state)


@action('go_to_url')
def go_to_url(tab_state: TabViewState,
              url: str,
              update_nav_history: bool = True):

    parsed_url = Url(url)
    page = pamet.page(parsed_url.get_page_id())

    if not page:
        tab_state.page_view_state = None
        fsm.update_state(tab_state)
        return

    page_state = get_or_create_page_view(tab_state, page)

    # Apply the anchor if any
    anchor = parsed_url.get_anchor()
    height = DEFAULT_EYE_HEIGHT
    center = Point2D()
    if anchor:
        height = page_state.viewport_height
        if isinstance(anchor, Note):
            center = anchor.rect().center()
            note_view_state = page_state.view_state_for_note_own_id(
                anchor.own_id)
            map_page_actions.clear_child_selection(page_state)
            map_page_actions.update_child_selections(page_state,
                                                     {note_view_state: True})
        elif isinstance(anchor, tuple):
            height, center = anchor
        else:
            raise Exception

    # Close any note or page edit windows
    if tab_state.note_edit_view_state:
        note_actions.abort_editing_note(tab_state)
    if tab_state.page_edit_view_state:
        close_page_properties(tab_state)

    page_state = get_or_create_page_view(tab_state, page)
    set_page_view_state(tab_state, page_state)

    if height:
        page_state.viewport_height = height
    if center:
        page_state.viewport_center = center
    if height or center:
        fsm.update_state(page_state)

    if not update_nav_history:
        return

    # Add the url to the navigation history
    nav_history = tab_state.navigation_history

    # If the page has changed
    last_page_id = None
    if nav_history:
        old_url = Url(nav_history[-1])
        last_page_id = old_url.get_page_id()

    if last_page_id != page.id:
        # Append the url to the nav history
        nav_history.append(str(page_state.page_url()))

    # Keep the nav history size in check
    if len(nav_history) > MAX_NAVIGATION_HISTORY:
        nav_history.pop(0)

    tab_state.set_navigation_index(len(nav_history) - 1)
    fsm.update_state(tab_state)


@action('page.create_and_link_page')
def create_new_page(tab_state: TabViewState, mouse_position: Point2D):
    # This action assumes there's a page opened
    current_page_state = tab_state.page_view_state
    current_page = pamet.page(current_page_state.page_id)
    new_name = generate_page_name()
    new_page = Page(name=new_name)
    pamet.insert_page(new_page)

    # Create a link to the new page at the mouse position or center
    # (in the old one)
    note_position = current_page_state.unproject_point(mouse_position)
    new_page_link = TextNote.in_page(current_page)
    new_page_link.url = new_page.url()
    new_page_link.x = note_position.x()
    new_page_link.y = note_position.y()
    pamet.insert_note(new_page_link)

    # Create a link to "old" page in the new one
    back_link_note = TextNote.in_page(new_page)
    back_link_note.url = current_page.url()
    rect = back_link_note.rect()
    rect.move_center(Point2D(0, 0))
    back_link_note.set_rect(rect)
    pamet.insert_note(back_link_note)

    # Open the page and open its properties (with the name focused for editing)
    go_to_url(tab_state, new_page.url())
    map_page_actions.open_page_properties(tab_state, focused_prop='name')


@action('navigation_back')
def navigation_back(tab_state: TabViewState):
    nav_history = tab_state.navigation_history
    if not nav_history:
        log.error('Navigation back called with no navigation history present')
        return

    tab_state.set_navigation_index(tab_state.current_nav_index - 1)
    fsm.update_state(tab_state)
    go_to_url(tab_state,
              nav_history[tab_state.current_nav_index],
              update_nav_history=False)


@action('navigation_forward')
def navigation_forward(tab_state: TabViewState):
    nav_history = tab_state.navigation_history
    if not nav_history:
        log.error(
            'Navigation forward called with no navigation history present')
        return

    tab_state.set_navigation_index(tab_state.current_nav_index + 1)
    fsm.update_state(tab_state)
    go_to_url(tab_state,
              nav_history[tab_state.current_nav_index],
              update_nav_history=False)


@action('navigation_toggle_last')
def navigation_toggle_last(tab_state):
    if tab_state.previous_nav_index is None:
        return

    nav_history = tab_state.navigation_history
    tab_state.set_navigation_index(tab_state.previous_nav_index)
    fsm.update_state(tab_state)
    go_to_url(tab_state,
              nav_history[tab_state.current_nav_index],
              update_nav_history=False)


@action('tab.close_page_properties')
def close_page_properties(tab_state: TabViewState):
    tab_state.page_edit_view_state = None
    fsm.update_state(tab_state)


@action('tab.open_global_search')
def open_global_search(tab_state: TabViewState):
    # If not present - create the search bar state
    # if not tab_state.search_bar_state:
    #     tab_state.search_bar_state = SearchBarWidgetState()
    #     fsm.add_state(tab_state.search_bar_state)

    tab_state.left_sidebar_state = SearchBarWidgetState()
    fsm.update_state(tab_state)


@action('tab.open_semantic_search')
def open_semantic_search(tab_state: TabViewState):
    # If not present - create the search bar state
    # if not tab_state.search_bar_state:
    #     tab_state.search_bar_state = SemanticSearchBarWidgetState()
    #     fsm.add_state(tab_state.search_bar_state)

    tab_state.left_sidebar_state = SemanticSearchBarWidgetState()
    fsm.update_state(tab_state)


@action('tab.close_left_sidebar')
def close_left_sidebar(tab_state: TabViewState):
    tab_state.left_sidebar_state = None
    fsm.update_state(tab_state)



from datetime import datetime, timedelta

import json
from pathlib import Path
import time
from typing import List

from fusion.libs.entity.change import Change

import pamet
from fusion.util import fake_time, current_time
from pamet.model.page import Page
from pamet.model.text_note import TextNote
from pamet.services.backup import FSStorageBackupService
from pamet.storage.file_system.repository import FSStorageRepository
from pamet.storage.pamet_in_memory_repo import PametInMemoryRepository

# Debug
PROCESS_INTERVAL = 5
PRUNE_INTERVAL = 15
BACKUP_INTERVAL = 10
PERMANENT_BACKUP_AGE = 30


def test_change_processing(tmp_path):
    backup_service = FSStorageBackupService(
        backup_folder=tmp_path, repository=PametInMemoryRepository())

    page = Page(name='test')
    change = Change.CREATE(page)
    backup_service.handle_change_set([change])
    backup_service.process_changes()

    # Assert result
    jsonl_data = backup_service.tmp_changeset_path(page.id).read_text()
    json_lines = jsonl_data.split('\n')
    assert json_lines[0] == json.dumps(change.as_safe_delta_dict(),
                                       ensure_ascii=False)


def change_and_backup(backup_service,
                      page,
                      time,
                      note_text: str = 'Test note'):
    note = TextNote.in_page(page)
    note.text = note_text
    with fake_time(time=time):
        change = pamet.insert_note(note, page)

        backup_service.handle_change_set([change])
        backup_service.process_changes()
        backup_service.backup_changed_pages()

    return change


def changes_are_identical(changeset1, changeset2):
    changeset1 = sorted(changeset1, key=lambda change: change.id)
    changeset1 = [change.as_safe_delta_dict() for change in changeset1]

    changeset2 = sorted(changeset2, key=lambda change: change.id)
    changeset2 = [change.as_safe_delta_dict() for change in changeset2]
    return changeset1 == changeset2


def test_backup(tmp_path):
    in_mem_repo = PametInMemoryRepository()
    backup_service = FSStorageBackupService(backup_folder=tmp_path,
                                            repository=in_mem_repo,
                                            record_all_changes=True)
    pamet.set_sync_repo(in_mem_repo)

    page = Page(name='test')
    change1 = pamet.insert_page(page)
    page_json1 = FSStorageRepository.serialize_page(page, [], [])

    backup_service.handle_change_set([change1])
    backup_service.process_changes()
    backup_service.backup_changed_pages()

    backups = list(backup_service.recent_backups_for_page(page.id))
    assert len(backups) == 1
    assert page_json1 == backups[0].read_text()

    note = TextNote.in_page(page)
    note.text = 'Test note'

    change2 = pamet.insert_note(note, page)
    page_json2 = FSStorageRepository.serialize_page(page, [note], [])

    backup_service.handle_change_set([change2])
    backup_service.process_changes()
    with fake_time(time=current_time() + timedelta(seconds=1)):
        backup_service.backup_changed_pages()

    backups = list(backup_service.recent_backups_for_page(page.id))
    assert len(backups) == 2
    assert page_json2 == backups[1].read_text()

    # changes_files = list(backup_service.changeset_paths_for_page(page.id))
    # assert len(changes_files) == 1
    # lines = changes_files[0].read_text().split('\n')
    # assert lines[0] == json.dumps(change.as_safe_delta_dict())
    changes = backup_service.get_changes()
    assert changes_are_identical(changes, [change1, change2])


def test_pruning(tmp_path):
    in_mem_repo = PametInMemoryRepository()
    backup_service = FSStorageBackupService(backup_folder=tmp_path,
                                            repository=in_mem_repo,
                                            record_all_changes=True)

    pamet.set_sync_repo(in_mem_repo)

    for_removal: List[Path] = []
    for_keeping: List[Path] = []
    change_obj_files = {}  # Change lists, by path

    # Craete the initial backup
    page = Page(name='test')
    page2 = Page(name='test2')  # To test page_id filtering for .get_changes()

    initial_backup_time = datetime(year=2099,
                                   month=7,
                                   day=30,
                                   hour=23,
                                   minute=1,
                                   second=0).astimezone()
    with fake_time(initial_backup_time):
        page1_change = pamet.insert_page(page)
        page2_change = pamet.insert_page(page2)
        backup_service.handle_change_set([page1_change, page2_change])
        backup_service.process_changes()
        backup_service.backup_changed_pages()

    for_keeping.append(
        backup_service.recent_backup_path(page.id, initial_backup_time))

    # Do x2 backups per period (last day, month, year). The older for each
    # Current time when pruning will be 2100.6.30 23:01:00

    # Last year, same week
    # This one should be deleted
    weekly_deleted_time = datetime.fromisocalendar(year=2100, week=10,
                                                   day=1).astimezone()
    for_removal.append(
        backup_service.recent_backup_path(page.id, weekly_deleted_time))
    change = change_and_backup(backup_service, page, weekly_deleted_time,
                               'Last year, same week, to be deleted')
    changes = [change]

    # This one should be kept, and the changes up to this point
    # should be merged
    weekly_kept_time = datetime.fromisocalendar(year=2100, week=10,
                                                day=2).astimezone()
    change = change_and_backup(backup_service, page, weekly_kept_time,
                               'Last year, same week, to be kept')
    for_keeping.append(
        backup_service.recent_backup_path(page.id, weekly_kept_time))
    changes.append(change)
    change_obj_files[weekly_kept_time] = changes

    # Last month, same day
    # This one should be deleted
    daily_deleted_time = datetime(year=2100,
                                  month=6,
                                  day=15,
                                  hour=2,
                                  minute=1,
                                  second=0).astimezone()
    change = change_and_backup(backup_service, page, daily_deleted_time,
                               'Last month, same day, to be deleted')
    for_removal.append(
        backup_service.recent_backup_path(page.id, daily_deleted_time))
    changes = [change]

    # This one should be kept, and the changes up to this point
    # should be merged
    daily_kept_time = datetime(year=2100,
                               month=6,
                               day=15,
                               hour=14,
                               minute=1,
                               second=0).astimezone()
    for_keeping.append(
        backup_service.recent_backup_path(page.id, daily_kept_time))
    change = change_and_backup(backup_service, page, daily_kept_time,
                               'Last month, same day, to be kept')
    changes.append(change)
    change_obj_files[daily_kept_time] = changes

    # Last day, same hour
    # This one should be deleted
    hourly_deleted_time = datetime(year=2100,
                                   month=6,
                                   day=30,
                                   hour=15,
                                   minute=1,
                                   second=0).astimezone()
    for_removal.append(
        backup_service.recent_backup_path(page.id, hourly_deleted_time))
    change = change_and_backup(backup_service, page, hourly_deleted_time,
                               'Last day, same hour, to be deleted')
    changes = [change]

    # This one should be kept
    hourly_kept_time = datetime(year=2100,
                                month=6,
                                day=30,
                                hour=15,
                                minute=20,
                                second=0).astimezone()
    for_keeping.append(
        backup_service.recent_backup_path(page.id, hourly_kept_time))
    change = change_and_backup(backup_service, page, hourly_kept_time,
                               'Last day, same hour, to be kept')
    changes.append(change)
    change_obj_files[hourly_kept_time] = changes

    # Do pruning
    with fake_time(
            datetime(year=2100, month=6, day=30, hour=23, minute=1,
                     second=0).astimezone()):
        backup_service.prune_page_backups(page.id)

    # Assert file counts and changes-file contents
    assert len(for_keeping) == len(
        backup_service.recent_backups_for_page(page.id))

    for backup_path in for_keeping:
        assert backup_path.exists()
    for backup_path in for_removal:
        assert not backup_path.exists()

    # Check if all changes are present and valid
    all_changes = []
    for change_path, changes in change_obj_files.items():
        all_changes.extend(changes)
    all_changes.extend([page1_change, page2_change])

    saved_changes = backup_service.get_changes()
    assert changes_are_identical(saved_changes, all_changes)

    # Check if the get_change filtering is working
    # Check for page_id filtering
    page2_changes = backup_service.get_changes(page_id=page2.id)
    assert changes_are_identical(page2_changes, [page2_change])

    # Chech for filtering by time
    all_changes = sorted(all_changes, key=lambda c: c.time)
    after_time = weekly_deleted_time
    before_time = hourly_deleted_time

    subset = list(
        filter(lambda c: after_time <= c.time < before_time, all_changes))
    retrieved = backup_service.get_changes(after_time=after_time,
                                           before_time=before_time)
    assert changes_are_identical(subset, retrieved)


def test_permanent_storage(tmp_path):
    in_mem_repo = PametInMemoryRepository()
    backup_service = FSStorageBackupService(backup_folder=tmp_path,
                                            repository=in_mem_repo)

    pamet.set_sync_repo(in_mem_repo)

    # Craete the initial backup
    page = Page(name='test')
    change = pamet.insert_page(page)

    initial_backup_time = datetime(year=2000, month=7, day=30).astimezone()
    with fake_time(initial_backup_time):
        backup_service.handle_change_set([change])
        backup_service.process_changes()
        backup_service.backup_changed_pages()

    backup_service.move_to_permanent_where_appropriate(page.id)

    assert not backup_service.recent_backups_for_page(page.id)
    assert backup_service.permanent_backup_path(page.id,
                                                initial_backup_time).exists()


def test_sheduler_and_worker(tmp_path):
    # Test the scheduler, worker thread and changes channel.
    # Just do a single backup and check it along with the backup/prune
    # timesetamps

    in_mem_repo = PametInMemoryRepository()
    pamet.set_sync_repo(in_mem_repo)

    backup_service = FSStorageBackupService(backup_folder=tmp_path,
                                            repository=in_mem_repo,
                                            process_interval=0.5,
                                            backup_interval=1,
                                            prune_interval=1.5)
    # Start the service
    backup_service.start()
    # time_started = time.time()
    assert backup_service.service_lock_path().exists()

    # Enter a change
    page = Page(name='test')
    change = pamet.insert_page(page)
    backup_service.handle_change_set([change])

    # Wait for the process_events
    time.sleep(0.6)
    assert backup_service.tmp_changeset_path(page.id).exists()

    time.sleep(1)
    assert bool(backup_service.last_backup_time(page.id))

    # time_requesting_stop = time.time()
    backup_service.stop()
    # time_stopped = time.time()
    assert not backup_service.service_lock_path().exists()

    assert backup_service.last_backup_timestamp_path().exists()
    assert backup_service.last_prune_timestamp_path().exists()
